from .datasets import EMDataset
import numpy as np
import pandas as pd
from scipy.spatial import cKDTree as kdtree
from scipy.interpolate import NearestNDInterpolator
from .utils import inverse_distance_interpolation

def find_locations_in_distance(xy_input, xy_output, distance=100.0):
    """
    Find indicies of locations of xy_output within a given separation distance
    from locations of xy_input.

    Parameters
    ----------

    xy_input: (*,2) array_like
        Input locations.
    xy_output: (*,2) array_like
        Ouput Locations where the indicies of the locations are sought.
    distance: float
        Separation distance used as a threshold

    Returns
    -------
    pts : (*,2) ndarray, float
        Sought locations.
    inds: (*,) ndarray, integer
        Sought indicies.
    """
    tree = kdtree(xy_output)
    out = tree.query_ball_point(xy_input, distance)
    temp = np.unique(out)
    inds = []
    for ind in temp:
        if ind != []:
            inds.append(ind)
    if len(inds) == 0:
        return None, None
    inds = np.unique(np.hstack(inds))
    pts = xy_output[inds, :]
    return pts, inds


def find_closest_locations(xy_input, xy_output):
    """
    Find indicies of the closest locations of xy_output from from locations of xy_input.

    Parameters
    ----------

    xy_input: (*,2) array_like
        Input locations.
    xy_output: (*,2) array_like
        Ouput Locations where the indicies of the locations are sought.

    Returns
    -------
    d : (*,) ndarray, float
        Closest distance.
    inds: (*,) ndarray, integer
        Sought indicies.
    """
    tree = kdtree(xy_output)
    d, inds = tree.query(xy_input)
    return d, inds


# TODO:
# Calcuate fraction for each lithologic unit
# Need to simplify this use volume avearging
def compute_fraction_for_aem_layer(hz, lith_data, unique_code):
    """
    Compute fraction of lithology in AEM layers

    Parameters
    ----------

    hz: (n_layer,) array_like
        Thickness of the AEM layers
    lith_data: pandas DataFrame including ['From', 'To', 'Code']
        Lithology logs
    unique_code: array_like
        uniuqe lithology code; n_code = unique.size

    Returns
    -------
    fraction : (n_layer, n_code) ndarray, float
        Fractoin of each lithologic code (or unit)
    """
    n_code = unique_code.size
    z_top = lith_data.From.values
    z_bottom = lith_data.To.values
    z = np.r_[z_top, z_bottom[-1]]
    code = lith_data.Code.values
    zmin = z_top.min()
    zmax = z_bottom.max()
    depth = np.r_[0.0, np.cumsum(hz)][:]
    z_aem_top = depth[:-1]
    z_aem_bottom = depth[1:]

    # assume lithology log always start with zero depth
    # TODO: at the moment, the bottom aem layer, which overlaps with a portion of the driller's log
    # is ignored.
    n_layer = (z_aem_bottom < zmax).sum()
    fraction = np.ones((hz.size, n_code)) * np.nan

    for i_layer in range(n_layer):
        inds_in = np.argwhere(
            np.logical_and(z >= z_aem_top[i_layer], z <= z_aem_bottom[i_layer])
        ).flatten()
        dx_aem = z_aem_bottom[i_layer] - z_aem_top[i_layer]
        if inds_in.sum() != 0:
            z_in = z[inds_in]
            dx_in = np.diff(z_in)
            code_in = code[inds_in[:-1]]
            if i_layer == 0:
                inds_bottom = inds_in[-1] + 1
                inds = np.r_[inds_in, inds_bottom]
                z_tmp = z[inds]
                dx_bottom = z_aem_bottom[i_layer] - z[inds_bottom - 1]
                dx = np.r_[dx_in, dx_bottom]
                code_bottom = code[inds_bottom - 1]
                code_tmp = np.r_[code_in, code_bottom]
            else:
                inds_bottom = inds_in[-1] + 1
                inds_top = inds_in[0] - 1
                inds = np.r_[inds_top, inds_in, inds_bottom]
                z_tmp = z[inds]
                dx_top = z[inds_top + 1] - z_aem_top[i_layer]
                dx_bottom = z_aem_bottom[i_layer] - z[inds_bottom - 1]
                dx = np.r_[dx_top, dx_in, dx_bottom]
                code_bottom = code[inds_bottom - 1]
                code_top = code[inds_top]
                code_tmp = np.r_[code_top, code_in, code_bottom]
        else:
            inds_top = np.argmin(abs(z - z_aem_top[i_layer]))
            inds_bottom = inds_top + 1
            inds = np.r_[inds_top, inds_bottom]
            z_tmp = z[inds]
            dx = np.r_[dx_aem]
            #     print (code[inds_top])
            code_tmp = np.r_[code[inds_top]]
        for i_code, unique_code_tmp in enumerate(unique_code):
            fraction[i_layer, i_code] = dx[code_tmp == unique_code_tmp].sum() / dx_aem
    return fraction



def generate_water_level_map(water_level_df: pd.DataFrame, em_data: EMDataset):
    xy_wse = water_level_df[['X', 'Y']].values
    wse = -water_level_df['TSZ'].values
    
    # nearest interpolation
    f_int_wse = NearestNDInterpolator(xy_wse, wse)
    wse_em = f_int_wse(em_data.xy)

    x, y, wse_idw = inverse_distance_interpolation(
        em_data.xy, wse_em, 
        dx=500, dy=500,
        max_distance=1000., k_nearest_points=200, power=0,
        x_pad=2000, y_pad=2000,
    )
    
    X, Y = np.meshgrid(x, y)
    # values_tmp = wse_idw[~wse_idw.mask]
    # x_vec_tmp = X[~wse_idw.mask]
    # y_vec_tmp = Y[~wse_idw.mask]
    
    df_water_table_grid = pd.DataFrame(data=np.c_[X.flatten(), Y.flatten(), wse_idw.data.flatten()], columns=['x', 'y', 'water_table'])
    
    return dict(
        x=x,
        y=y,
        X=X,
        Y=X,
        wse_idw=wse_idw,
        wse_em=wse_em,
        df=df_water_table_grid,
        nn_interpolator=f_int_wse
    )

def compute_colocations(distance_threshold: int, em_data: EMDataset, df_lithology_collar: pd.DataFrame):
    
    xy_lithology = df_lithology_collar[['X', 'Y']].values
    
    # find the well locations that are within the distance_threshold of any em suruvey location
    xy_lithology_colocated, inds_driller = find_locations_in_distance(em_data.xy, xy_lithology, distance=distance_threshold)

    # use each well location to lookup the cloest em survey location
    d_aem_colocated, inds_aem_colocated = find_closest_locations(xy_lithology_colocated, em_data.xy)

    # get the subset of co-located survey locations
    xy_aem_colocated = em_data.xy[inds_aem_colocated,:]

    # get the subset of co-located wells
    _, inds_lithology_colocated = find_closest_locations(xy_aem_colocated, xy_lithology)
    df_lithology_collar_colocated = df_lithology_collar.loc[inds_lithology_colocated]

    # get names of the colocated wells
    well_names_colocated = df_lithology_collar_colocated['WellID'].values

    # there should always be the same number of co-ocated wells and survey locations
    assert inds_aem_colocated.size == inds_driller.size
    
    n_colocated = inds_aem_colocated.size
    mean_separation_distance = d_aem_colocated.mean()
    
    return dict(
        n_colocated=n_colocated,
        mean_separation_distance=mean_separation_distance,
        xy_em=xy_aem_colocated,
        lithology_collar=df_lithology_collar_colocated,
        well_names=well_names_colocated,
        inds_em=inds_aem_colocated,
        inds_lithology=inds_lithology_colocated
    )


from collections import namedtuple
import numpy as np
import pandas as pd

class EMDataset():
    filename: str;
    df: pd.DataFrame;
    
    def __init__(self, csv_filename: str):
        self.filename = csv_filename;
        self.df = pd.read_csv(self.filename, dtype={'LINE_NO': 'Int32', 'RECORD': 'Int32'}).sort_values('RECORD')
        
    @property
    def header(self):
        return list(self.df.columns)
    
    @property
    def columns(self):
        return namedtuple("columns",["thk","rho"])(
                np.array(self.header)[['THK' in h for h in self.header]],
                np.array(self.header)[['RHO' in h for h in self.header]]
        )
    
    @property
    def names(self):
        return namedtuple("names",["thk","rho"])(
            self.columns.thk[:int(len(self.columns.thk)/2)],
            self.columns.rho[:int(len(self.columns.thk)/2)+1]
        )
    
    @property
    def line(self):
        return self.df['LINE_NO'].values
    
    @property
    def timestamps(self):
        return self.df['TIMESTAMP'].values
    
    @property
    def topography(self):
        return self.df[['UTMX', 'UTMY', 'ELEVATION']].values[:, :]
    
    @property
    def hz(self):
        hz = np.unique(self.df[self.names.thk].values)
        return np.r_[hz, hz[-1]]
    
    @property
    def resistivity(self):
        return self.df[self.names.rho].values[:,:]

    @property
    def num_soundings(self):
        return self.df[self.names.rho].values[:,:].shape[0]

    @property
    def xy(self):
        return self.df[['UTMX', 'UTMY']].values

    @property
    def lines_xy(self):
        lines = []
        em_lines = dict()
        for g, data in self.df[['LINE_NO','RECORD','UTMX', 'UTMY']].groupby('LINE_NO'):
            lines.append(g)
            em_lines[g] = data[['UTMX', 'UTMY', 'RECORD']].values
        return lines, em_lines
    
    @property
    def num_layers(self):
        return self.hz.size
    
    def get_resistivity_by_line(self, line_number: int):
        records = self.df[self.df['LINE_NO'] == line_number][['UTMX','UTMY', *self.names.rho]]
        values = records.values[:, :]
        rho = values[:, 2:]
        xy = values[:, :2]
        delta = np.concatenate([[0],(np.diff(values[:, 0])**2 + np.diff(values[:, 1])**2)**.5])
        return rho, delta, xy



from collections import namedtuple
from traitlets import HasTraits, TraitType
from ipywidgets import Output
import re
import os
import requests

RegisteredWidget = namedtuple('RegisteredWidget', ['name','widget'])

class AppState(HasTraits):
    widgets: dict[str, any]
    
    @property
    def traits(self):
        return list(self.__dict__.keys())

    def to_dict(self):
        return self.__dict__
    
    def register_stateful_widget(self, widget: any, trait_name: str, traitlet: TraitType):
        if not hasattr(self, 'widgets'):
            self.widgets = {}
        self.add_traits(**{trait_name: traitlet})
        self.widgets[trait_name] = widget
        self.widgets[trait_name].observe(lambda _: self.update_from_ui(), names=["value"])
    
    def register_stateful_property(self, trait_name: str, traitlet: TraitType):
        self.add_traits(**{trait_name: traitlet})
    
    def register_widget_observer(self, observes: str, observer):
        self.widgets[observes].observe(observer, names=["value"])
        
    def update_from_ui(self):
        for name in self.widgets.keys():
            setattr(self, name, self.widgets[name].value)

    def restore_ui(self):
        for name in self.widgets.keys():
            self.widgets[name].value = self[name]
    
    def observe_trait(self, trait_name: str, fn):
        def observer(state):
            if state["name"] == trait_name:
                fn(state)
        
        return self.observe(observer)
        
    
    @property
    def outlet(self):
        out = Output()
        def print_state(s):
            with out:
                print(s)
        self.observe(print_state)
        with out:
            print("AppState Outlet:")
        return out

def validate_upload_specifications(list_of_things_to_upload):
    expected_keys = ['local_filepath', 'name', 'description', 'content_type']
    for upload in list_of_things_to_upload:
        if not all(key in upload for key in expected_keys):
            raise ValueError(f"missing key in {upload} - expected all of {expected_keys}")
        if not os.path.exists(upload["local_filepath"]):
            raise ValueError(f"missing local file: {upload['local_filepath']}")

def upload_a_file(filepath: str, name: str, content_type: str):
    # 1. get a signed url to upload to
    url = f'{os.environ["RECHARGE_API_ENDPOINT"]}/upload'
    filesize = os.stat(filepath).st_size
    r = requests.post(
        url,
        headers={
            "content-type": 'application/json',
            "authorization": f"Bearer {os.environ['RECHARGE_API_TOKEN']}"
        },
        json = {
        'filename': re.sub("[_/.]", '-', name),
        'content_type': content_type,
        'size': filesize
    },
    )
    
    if r.status_code >= 299:
        raise ValueError(f"Error {r.status_code} - {r.text}")
    
    upload_info = r.json()
    
    # 2. Upload the file.
    r = requests.put(
        upload_info["url"],
        headers={
            "content-type": content_type,
            "Content-Range": "bytes 0-" + str(filesize - 1) + "/" + str(filesize)
        },
        data=open(filepath, 'rb')
    )
    
    if r.status_code > 299:
        raise ValueError(f"Error {r.status_code} - {r.text}")
    
    return {"id": upload_info["id"], "size": filesize}
    
def add_exports_to_simulation(simulationId, list_of_exports):
    url = f'{os.environ["RECHARGE_API_ENDPOINT"]}/simulations/{simulationId}/exports'
    r= requests.patch(
        url,
        headers={
            "content-type": 'application/json',
            "authorization": f"Bearer {os.environ['RECHARGE_API_TOKEN']}"
        },
        json={
            "items": list_of_exports
        }
    )

    if r.status_code >= 299:
        raise ValueError(f"Error {r.status_code} - {r.text}")

def upload_and_export_all(list_of_things_to_upload):
    print(f"Uploading {len(list_of_things_to_upload)} outputs")
    errors = []
    exports_to_add = []
    for upload in list_of_things_to_upload:
        print(f"Uploading {upload['name']}...", end="", flush=True)
        try:
            upload_info = upload_a_file(upload["local_filepath"], upload["name"], upload["content_type"])
            exports_to_add.append({**upload, **upload_info})
            print("DONE")
        except Exception as err:
            print("ERROR!")
            errors.append(err)
    
    print("File uploads completed!")
    
    if len(errors) > 0:
        print(f"{len(errors)} errors encountered")
        for n, err in enumerate(errors):
            print(f"{n}: {err}")
    
    print("Upload success!")
    print("Updating simulation...")    
    add_exports_to_simulation(os.environ["SIMULATION_ID"], exports_to_add)
    print("Update completed!")